VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "CProcess"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' *************************************************************************
'  Copyright 2000-01 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' *************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' *************************************************************************
'  See also:
'    Q175030 - HOWTO: Enumerate Applications in Win32
'    http://support.microsoft.com/support/kb/articles/Q175/0/30.asp
'    Q296526 - INFO: Collecting Performance Data Using PDH APIs from Visual Basic
'    http://support.microsoft.com/support/kb/articles/q296/5/26.asp
'    Q192986 - Modlist.exe Shows How to Enumerate Processes and Modules
'    http://support.microsoft.com/support/kb/articles/q192/9/86.asp
'    Q187913 - HOWTO: List Running Processes
'    http://support.microsoft.com/support/kb/articles/q187/9/13.asp
' *************************************************************************
Option Explicit

' Win32 process-related APIs
Private Declare Function OpenProcess Lib "kernel32" (ByVal dwDesiredAccess As Long, ByVal bInheritHandle As Long, ByVal dwProcessID As Long) As Long
Private Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long

' Other useful APIs
Private Declare Function SHGetFileInfo Lib "shell32.dll" Alias "SHGetFileInfoA" (ByVal pszPath As String, ByVal dwFileAttributes As Long, psfi As SHFILEINFO, ByVal cbFileInfo As Long, ByVal uFlags As Long) As Long

' PSAPI Declares...
Private Declare Function EnumProcessModules Lib "PSAPI" (ByVal hProcess As Long, lphModule As Any, cb As Long, lpcbNeeded As Long) As Long
Private Declare Function GetModuleBaseName Lib "PSAPI" Alias "GetModuleBaseNameA" (ByVal hProcess As Long, ByVal hModule As Long, ByVal lpBasename As String, nSize As Long) As Long
Private Declare Function GetModuleFileNameEx Lib "PSAPI" Alias "GetModuleFileNameExA" (ByVal hProcess As Long, ByVal hModule As Long, ByVal lpFileName As String, nSize As Long) As Long
Private Declare Function GetModuleInformation Lib "PSAPI" (ByVal hProcess As Long, ByVal hModule As Long, LPMODULEINFO As MODULEINFO, cb As Long) As Long
Private Declare Function GetProcessMemoryInfo Lib "PSAPI" (ByVal Process As Long, ppsmemCounters As PROCESS_MEMORY_COUNTERS, cb As Long) As Long

' Used by GetModuleInformation
Private Type MODULEINFO
   lpBaseOfDll As Long
   SizeOfImage As Long
   EntryPoint As Long
End Type

' Used by GetProcessMemoryInfo
Private Type PROCESS_MEMORY_COUNTERS
   cb As Long
   PageFaultCount As Long
   PeakWorkingSetSize As Long
   WorkingSetSize As Long
   QuotaPeakPagedPoolUsage As Long
   QuotaPagedPoolUsage As Long
   QuotaPeakNonPagedPoolUsage As Long
   QuotaNonPagedPoolUsage As Long
   PageFileUsage As Long
   PeakPageFileUsage As Long
End Type

' PSAPI is only available in NT!
Private Declare Function GetVersionEx Lib "kernel32" Alias "GetVersionExA" (lpVersionInformation As OSVERSIONINFO) As Long
Private Type OSVERSIONINFO
   dwOSVersionInfoSize As Long
   dwMajorVersion As Long
   dwMinorVersion As Long
   dwBuildNumber As Long
   dwPlatformId As Long
   szCSDVersion As String * 128
End Type
Private Const VER_PLATFORM_WIN32s = 0
Private Const VER_PLATFORM_WIN32_WINDOWS = 1
Private Const VER_PLATFORM_WIN32_NT = 2

' Maximum path length (without special handling) in NT
Private Const MAX_PATH As Long = 260&

' Used by the OpenProcess API call
Private Const PROCESS_ALL_ACCESS As Long = &H1F0FFF
Private Const PROCESS_CREATE_PROCESS As Long = &H80
Private Const PROCESS_CREATE_THREAD As Long = &H2
Private Const PROCESS_DUP_HANDLE As Long = &H40
Private Const PROCESS_HEAP_ENTRY_BUSY As Long = &H4
Private Const PROCESS_HEAP_ENTRY_DDESHARE As Long = &H20
Private Const PROCESS_HEAP_ENTRY_MOVEABLE As Long = &H10
Private Const PROCESS_HEAP_REGION As Long = &H1
Private Const PROCESS_HEAP_UNCOMMITTED_RANGE As Long = &H2
Private Const PROCESS_QUERY_INFORMATION As Long = &H400
Private Const PROCESS_SET_INFORMATION As Long = &H200
Private Const PROCESS_SET_QUOTA As Long = &H100
Private Const PROCESS_TERMINATE As Long = &H1
Private Const PROCESS_VM_OPERATION As Long = &H8
Private Const PROCESS_VM_READ As Long = &H10
Private Const PROCESS_VM_WRITE As Long = &H20

' SHGetFileInfo constants.
Private Const SHGFI_ICON = &H100                         '  get icon
Private Const SHGFI_DISPLAYNAME = &H200                  '  get display name
Private Const SHGFI_TYPENAME = &H400                     '  get type name
Private Const SHGFI_ATTRIBUTES = &H800                   '  get attributes
Private Const SHGFI_ICONLOCATION = &H1000                '  get icon location
Private Const SHGFI_EXETYPE = &H2000                     '  return exe type
Private Const SHGFI_SYSICONINDEX = &H4000                '  get system icon index
Private Const SHGFI_LINKOVERLAY = &H8000                 '  put a link overlay on icon
Private Const SHGFI_SELECTED = &H10000                   '  show icon in selected state
Private Const SHGFI_LARGEICON = &H0                      '  get large icon
Private Const SHGFI_SMALLICON = &H1                      '  get small icon
Private Const SHGFI_OPENICON = &H2                       '  get open icon
Private Const SHGFI_SHELLICONSIZE = &H4                  '  get shell size icon
Private Const SHGFI_PIDL = &H8                           '  pszPath is a pidl
Private Const SHGFI_USEFILEATTRIBUTES = &H10             '  use passed dwFileAttribute

' SHGetFileInfo structures.
Private Type SHFILEINFO
   hIcon As Long                       '  out: icon
   iIcon As Long                       '  out: icon index
   dwAttributes As Long                '  out: SFGAO_ flags
   szDisplayName As String * MAX_PATH  '  out: display name (or path)
   szTypeName As String * 80           '  out: type name
End Type

' Used to get API error descriptions
Private Declare Function FormatMessage Lib "kernel32" Alias "FormatMessageA" _
   (ByVal dwFlags As Long, lpSource As Any, ByVal dwMessageId As Long, _
   ByVal dwLanguageId As Long, ByVal lpBuffer As String, ByVal nSize As Long, _
   Arguments As Long) As Long
Private Const FORMAT_MESSAGE_FROM_SYSTEM As Long = &H1000

' NTVDM modes:
Public Enum VdmTypes
   tlNotVdm = 0
   tlVdmWin16 = 1
   tlVdmDos = 2
End Enum

' Defaults
Private Const defHungTimeout = 500 'milliseconds
Private Const defDialogTimeout = 5000 'milliseconds

' Member variables
Private m_RunningNT As Boolean
Private m_PID As Long
Private m_hProcess As Long
Private m_Mods() As Long
Private m_ModCount As Long
Private m_PMC As PROCESS_MEMORY_COUNTERS
Private m_PMC_Empty As PROCESS_MEMORY_COUNTERS
Private m_Valid As Boolean

' Member constants
Private Const nMaxMods As Long = 1024
Private Const nBaseModule As Long = 1
Private Const BytesPerLong As Long = 4
Private Const NullProcess As Long = -1&

' ************************************************
'  Initialize/Terminate
' ************************************************
Private Sub Class_Initialize()
   Dim os As OSVERSIONINFO
   
   ' Make sure we're in NT.
   os.dwOSVersionInfoSize = Len(os)
   Call GetVersionEx(os)
   m_RunningNT = (os.dwPlatformId = VER_PLATFORM_WIN32_NT)
   
   ' Provide lots of space initially for module
   ' enumeration, and increase as needs grow.
   ReDim m_Mods(nBaseModule To nMaxMods) As Long
   
   ' Initialize some vars
   m_PMC.cb = Len(m_PMC)
   m_PMC_Empty.cb = Len(m_PMC_Empty)
   m_PID = NullProcess
End Sub

Private Sub Class_Terminate()
   ' Close open process
   If m_hProcess Then Call CloseHandle(m_hProcess)
End Sub

' ************************************************
'  Public Properties: Read-only
' ************************************************
Public Property Get DisplayName() As String
   Dim sRet As String
   Dim sfi As SHFILEINFO
   
   If m_hProcess Then
      ' Get fully-qualified path to base module.
      sRet = Me.ModuleName(nBaseModule, True)
      
      ' Trim crap off return on some files?
      If Left$(sRet, 4) = "\??\" And Len(sRet) > 4 Then
         sRet = Mid$(sRet, 5)
      End If
      
      ' Gussy it up for display, returning
      ' ModuleName on failure of SHFGI.
      If SHGetFileInfo(sRet, 0&, sfi, Len(sfi), SHGFI_DISPLAYNAME) Then
         sRet = TrimNull(sfi.szDisplayName)
      End If
      DisplayName = sRet
   End If
End Property

Public Property Get hProcess() As Long
   hProcess = m_hProcess
End Property

Public Property Get ModuleBaseAddress(Optional ByVal Module As Long = 1) As Long
   Dim mi As MODULEINFO
   Dim nRet As Long
   
   If m_hProcess Then
      ' Retrieve load address of the module.
      nRet = GetModuleInformation(m_hProcess, m_Mods(Module), mi, Len(mi))
      If nRet Then
         ModuleBaseAddress = mi.lpBaseOfDll
      Else
         ApiErrorDump Err.LastDllError, "GetModuleInformation"
      End If
   End If
End Property

Public Property Get ModuleCount() As Long
   ' Return cached count of loaded modules.
   ModuleCount = m_ModCount
End Property

Public Property Get ModuleEntryPoint(Optional ByVal Module As Long = 1) As Long
   Dim mi As MODULEINFO
   Dim nRet As Long
   
   If m_hProcess Then
      ' Retrieve entry point of the module.
      nRet = GetModuleInformation(m_hProcess, m_Mods(Module), mi, Len(mi))
      If nRet Then
         ModuleEntryPoint = mi.EntryPoint
      Else
         ApiErrorDump Err.LastDllError, "GetModuleInformation"
      End If
   End If
End Property

Public Property Get ModuleImageSize(Optional ByVal Module As Long = 1) As Long
   Dim mi As MODULEINFO
   Dim nRet As Long
   
   If m_hProcess Then
      ' Retrieve size, in bytes, of the linear space
      ' that the module occupies.
      nRet = GetModuleInformation(m_hProcess, m_Mods(Module), mi, Len(mi))
      If nRet Then
         ModuleImageSize = mi.SizeOfImage
      Else
         ApiErrorDump Err.LastDllError, "GetModuleInformation"
      End If
   End If
End Property

Public Property Get ModuleName(Optional ByVal Module As Long = 1, Optional ByVal FullName As Boolean = False) As String
   Dim sRet As String
   Dim nRet As Long
   
   If m_hProcess Then
      ' Make sure request is for legit module.
      If Module < nBaseModule Or Module > m_ModCount Then
         Module = nBaseModule
      End If
      
      ' Snag module name, fully qualified if asked.
      sRet = String$(MAX_PATH, 0)
      If FullName Then
         nRet = GetModuleFileNameEx(m_hProcess, m_Mods(Module), sRet, MAX_PATH)
      Else
         nRet = GetModuleBaseName(m_hProcess, m_Mods(Module), sRet, MAX_PATH)
      End If
      
      ' Return results or report error
      If nRet Then
         ModuleName = Left$(sRet, nRet)
      Else
         ' Houston, we have a problem.
         If FullName Then
            'ApiErrorDump Err.LastDllError, "GetModuleFileNameEx"
         Else
            'ApiErrorDump Err.LastDllError, "GetModuleBaseName"
         End If
      End If
   Else
      ' May have system idle process
      If m_PID = 0 Then
         ModuleName = "System Idle Process"
      End If
   End If
End Property

Public Property Get PageFaultCount() As Long
   PageFaultCount = m_PMC.PageFaultCount
End Property

Public Property Get PageFileUsage() As Long
   PageFileUsage = m_PMC.PageFileUsage
End Property

Public Property Get PeakPageFileUsage() As Long
   PeakPageFileUsage = m_PMC.PeakPageFileUsage
End Property

Public Property Get PeakWorkingSetSize() As Long
   PeakWorkingSetSize = m_PMC.PeakWorkingSetSize
End Property

Public Property Get QuotaNonPagedPoolUsage() As Long
   QuotaNonPagedPoolUsage = m_PMC.QuotaNonPagedPoolUsage
End Property

Public Property Get QuotaPagedPoolUsage() As Long
   QuotaPagedPoolUsage = m_PMC.QuotaPagedPoolUsage
End Property

Public Property Get QuotaPeakNonPagedPoolUsage() As Long
   QuotaPeakNonPagedPoolUsage = m_PMC.QuotaPeakNonPagedPoolUsage
End Property

Public Property Get QuotaPeakPagedPoolUsage() As Long
   QuotaPeakPagedPoolUsage = m_PMC.QuotaPeakPagedPoolUsage
End Property

Public Property Get VDM() As VdmTypes
   Dim nRet As Long
   
   ' Assume this isn't in a VDM.
   nRet = tlNotVdm
   
   ' No VDMs in Win9x.
   If m_RunningNT Then
      ' Fire off enumeration of all VDMs.
      If IsProcVdmWin16(m_PID) Then
         nRet = tlVdmWin16
      ElseIf IsProcVdmDOS(m_PID) Then
         nRet = tlVdmDos
      Else
         nRet = tlNotVdm
      End If
   End If
   
   ' Return results
   VDM = nRet
End Property

Public Property Get WorkingSetSize() As Long
   WorkingSetSize = m_PMC.WorkingSetSize
End Property

' ************************************************
'  Public Properties: Read/Write
' ************************************************
Public Property Get ProcessID() As Long
   ProcessID = m_PID
End Property

Public Property Let ProcessID(ByVal NewVal As Long)
   Const opFlags1 As Long = PROCESS_ALL_ACCESS
   Const opFlags2 As Long = PROCESS_QUERY_INFORMATION Or PROCESS_VM_READ
   Const opFlags3 As Long = PROCESS_QUERY_INFORMATION
   
   ' Check if new PID is different from cached.
   If NewVal <> m_PID Then
      ' Make sure we're running in NT!
      If m_RunningNT Then
         ' Close open process.
         If m_hProcess Then
            Call CloseHandle(m_hProcess)
            m_hProcess = 0
         End If
         
         ' Clear cached vars
         m_PMC = m_PMC_Empty
         
         ' Open new PID to query properties.
         ' Try successively more restrictive
         ' access rights until successful.
         m_PID = NewVal
         m_hProcess = OpenProcess(opFlags1, False, m_PID)
         If m_hProcess = 0 Then
            m_hProcess = OpenProcess(opFlags2, False, m_PID)
            If m_hProcess = 0 Then
               m_hProcess = OpenProcess(opFlags3, False, m_PID)
            End If
         End If

         ' Deal with problem processes...
         If m_hProcess = 0 Then
            ' Non-fatal and actually expected.  From MSDN:
            ' "For example, OpenProcess fails for the Idle
            ' and CSRSS processes because their access
            ' restrictions prevent user-level code from
            ' opening them."
            Select Case Err.LastDllError
               Case 5  ' Access is denied.
                  ' Probably CSRSS or System
               Case 87
                  ' Probably "System Idle Process"
               Case Else
                  ApiErrorDump Err.LastDllError, "OpenProcess"
            End Select
         End If
      End If
   End If
   
   ' Refresh process properties.
   Call Refresh
End Property

' ************************************************
'  Public Methods
' ************************************************
Public Function FormatBytes(ByVal Size As Long) As String
   Const KB As Long = 1024
   ' Return size in raw kilobytes.
   If Size >= 0 Then
      FormatBytes = Format$(Size \ KB, "#,##0") & " K"
   Else
      FormatBytes = "(" & Format$(-Size \ KB, "#,##0") & ") K"
   End If
End Function

Public Function FormatBytesEx(ByVal Size As Long, Optional ByVal Extended As Boolean = False) As String
   Dim sRet As String
   Const KB As Long = 1024
   Const MB As Long = KB * KB
   
   ' Return size of file in kilobytes.
   If Size < KB Then
      sRet = Format(Size, "#,##0") & " bytes"
   Else
      Select Case Size \ KB
         Case Is < 10
            sRet = Format(Size / KB, "0.00") & "KB"
         Case Is < 100
            sRet = Format(Size / KB, "0.0") & "KB"
         Case Is < 1000
            sRet = Format(Size / KB, "0") & "KB"
         Case Is < 10000
            sRet = Format(Size / MB, "0.00") & "MB"
         Case Is < 100000
            sRet = Format(Size / MB, "0.0") & "MB"
         Case Is < 1000000
            sRet = Format(Size / MB, "0") & "MB"
         Case Is < 10000000
            sRet = Format(Size / MB / KB, "0.00") & "GB"
      End Select
      
      ' Optionally return actual bytes as well
      If Extended Then
         sRet = sRet & " (" & Format(Size, "#,##0") & " bytes)"
      End If
   End If
   FormatBytesEx = sRet
End Function

Public Sub Refresh()
   Dim BytesNeeded As Long
   Dim BufferSize As Long
   Dim nRet As Long
      
   If m_hProcess Then
      ' Get list of process modules.
      BufferSize = BytesPerLong * UBound(m_Mods)
      nRet = EnumProcessModules(m_hProcess, m_Mods(1), BufferSize, BytesNeeded)
      If nRet = False Then
         ' Check to see if we need to allocate
         ' more space for results.
         If BytesNeeded > BufferSize Then
            ReDim m_Mods(nBaseModule To BytesNeeded \ BytesPerLong) As Long
            BufferSize = BytesNeeded
            nRet = EnumProcessModules(m_hProcess, m_Mods(1), BufferSize, BytesNeeded)
         End If
         
         ' Test for major failure, and bail if need be.
         If nRet = False Then
            m_ModCount = 0
            ' ApiErrorDump Err.LastDllError, "EnumProcessModules"
            Exit Sub
         End If
      End If
            
      ' Cache number of modules in this process.
      m_ModCount = BytesNeeded \ BytesPerLong
      
      ' Update process memory counters cache.
      ' Structure size set in Class_Initialize.
      Call GetProcessMemoryInfo(m_hProcess, m_PMC, m_PMC.cb)
      
      ' Check if this is a Virtual DOS Machine.
      If UCase$(Me.ModuleName()) = "NTVDM.EXE" Then
         Debug.Print "NTVDM!"; m_PID
      End If
      
   Else
      ' Check for system idle process
      If m_PID = 0 Then
         m_ModCount = 1
         
      End If
   End If
End Sub

Public Function Terminate(Optional DialogDelay As Long = defDialogTimeout) As Boolean
   ' Shut process down according to what type it is.
   ' Need to add separate routines if we want to shut
   ' down Win16 tasks within an NTVDM process.  Right
   ' now this just kills the entire NTVDM.
   Select Case Me.VDM()
      Case tlNotVdm, tlVdmWin16
         Call ProcessClose( _
            ProcessID:=m_PID, _
            TimeOut:=DialogDelay)
      'Case tlVdmWin16
      '   Call ProcessClose16(hWnd:=.hWnd, TimeOut:=DialogDelay)
      Case tlVdmDos
         Call ProcessCloseDOS(m_PID)
   End Select
End Function

' ************************************************
'  Private Methods
' ************************************************
Private Sub ApiErrorDump(ByVal ErrNum As Long, Optional ByVal Source As String = "")
   If Len(Source) Then
      Debug.Print Trim$(Source);
   Else
      Debug.Print "API";
   End If
   Debug.Print " error:"; ErrNum; " -- ";
   Debug.Print ApiErrorText(ErrNum);
End Sub

Private Function ApiErrorText(ByVal ErrNum As Long) As String
   Dim msg As String
   Dim nRet As Long

   msg = Space$(1024)
   nRet = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, ByVal 0&, ErrNum, 0&, msg, Len(msg), ByVal 0&)
   If nRet Then
      ApiErrorText = Left$(msg, nRet)
   Else
      ApiErrorText = "Error (" & ErrNum & ") not defined."
   End If
End Function

Private Function TrimNull(ByVal StrIn As String) As String
   Dim nul As Long
   
   ' Truncate input string at first null.
   ' If no nulls, perform ordinary Trim.
   nul = InStr(StrIn, vbNullChar)
   Select Case nul
      Case Is > 1
         TrimNull = Left(StrIn, nul - 1)
      Case 1
         TrimNull = ""
      Case 0
         TrimNull = Trim(StrIn)
   End Select
End Function

