VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "CProcesses"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' *************************************************************************
'  Copyright 2000-01 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' *************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' *************************************************************************
'  For the MS "answer" to this task, see also:
'    http://support.microsoft.com/support/kb/articles/Q187/9/13.asp
' *************************************************************************
Option Explicit

' PSAPI Declares...
Private Declare Function EnumProcesses Lib "PSAPI" (lpidProcess As Any, cb As Long, cbNeeded As Long) As Long
'Private Declare Function GetDeviceDriverBaseName Lib "PSAPI" Alias "GetDeviceDriverBaseNameA" (ByVal lpImageBase As Any, ByVal lpBasename As String, nSize As Long) As Boolean
'Private Declare Function GetDeviceDriverFileName Lib "PSAPI" Alias "GetDeviceDriverFileNameA" (ByVal lpImageBase As Long, ByVal lpFilename As String, nSize As Long) As Long
'Private Declare Function EnumDeviceDrivers Lib "PSAPI" (lpImageBase As Any, cb As Long, lpcbNeeded As Long) As Boolean

' PSAPI is only available in NT!
Private Declare Function GetVersionEx Lib "kernel32" Alias "GetVersionExA" (lpVersionInformation As OSVERSIONINFO) As Long
Private Type OSVERSIONINFO
   dwOSVersionInfoSize As Long
   dwMajorVersion As Long
   dwMinorVersion As Long
   dwBuildNumber As Long
   dwPlatformId As Long
   szCSDVersion As String * 128
End Type
Private Const VER_PLATFORM_WIN32s = 0
Private Const VER_PLATFORM_WIN32_WINDOWS = 1
Private Const VER_PLATFORM_WIN32_NT = 2

' Used to get API error descriptions
Private Declare Function FormatMessage Lib "kernel32" Alias "FormatMessageA" _
   (ByVal dwFlags As Long, lpSource As Any, ByVal dwMessageId As Long, _
   ByVal dwLanguageId As Long, ByVal lpBuffer As String, ByVal nSize As Long, _
   Arguments As Long) As Long
Private Const FORMAT_MESSAGE_FROM_SYSTEM As Long = &H1000

' Member variables
Private m_RunningNT As Boolean
Private m_PIDs() As Long
Private m_Procs As Collection
Private m_Count As Long

' Member constants
Private Const nMaxProcs As Long = 1024
Private Const BytesPerLong As Long = 4

' ************************************************
'  Initialize/Terminate
' ************************************************
Private Sub Class_Initialize()
   Dim os As OSVERSIONINFO
   
   ' Make sure we're in NT.
   os.dwOSVersionInfoSize = Len(os)
   Call GetVersionEx(os)
   m_RunningNT = (os.dwPlatformId = VER_PLATFORM_WIN32_NT)
   
   If m_RunningNT Then
      ' Provide lots of space initially for process
      ' enumeration, and increase as needs grow.
      ReDim m_PIDs(1 To nMaxProcs) As Long
      
      ' Update list of Process IDs (PIDs).
      Call Refresh
   End If
End Sub

Private Sub Class_Terminate()
   ' Shut down collection elements.
   Set m_Procs = Nothing
End Sub

' ************************************************
'  Public Properties: Read-only
' ************************************************
Public Property Get ProcessID(ByVal Index As Long) As Long
   If Index >= 1 And Index <= m_Count Then
      ProcessID = m_PIDs(Index)
   Else
      ' Potentially raise an "out of bounds" error.
   End If
End Property

Public Property Get RunningNT() As Boolean
   RunningNT = m_RunningNT
End Property

Public Property Get Valid(ByVal ProcessID As Long) As Boolean
   Dim i As Long
   ' Lots of potential strategies here. Could
   ' trap error if PID isn't valid, or iterate
   ' collection/array looking for PID.
   For i = 1 To m_Count
      If m_PIDs(i) = ProcessID Then
         Valid = True
         Exit For
      End If
   Next i
End Property

' ************************************************
'  Public Properties: Read/Write
' ************************************************

' ************************************************
'  Standard Public Collection Methods
' ************************************************
Public Function Add(ByVal CmdLine As String, Optional WinStyle As VbAppWinStyle = vbNormalFocus) As CProcess
   Dim PID As Long
   Dim proc As CProcess
   
   ' Range check on ExecMode.
   If WinStyle < vbHide Or WinStyle > vbMinimizedNoFocus Then
      WinStyle = vbMinimizedNoFocus
   End If
   
   ' Launch it, and add to collection.
   PID = Shell(CmdLine, WinStyle)
   If PID < 0 Or PID > 32 Then
      ' No apparent error starting app.
      Set proc = New CProcess
      proc.ProcessID = PID
      If proc.hProcess Then
         m_Procs.Add proc, CStr(proc.ProcessID)
         Set Add = m_Procs(CStr(PID))
      End If
   End If
End Function

Public Function Count() As Long
   Count = m_Count
End Function

Public Function Item(Optional ByVal Index As Long, Optional ByVal ID As Long) As CProcess
Attribute Item.VB_UserMemId = 0
   ' Check if we got a process ID.
   If ID <> 0 Then Index = IDtoIndex(ID)
   ' Return corresponding process
   If Index >= 1 And Index <= m_Count Then
      Set Item = m_Procs(Index)
   End If
End Function

Public Function NewEnum() As IUnknown
Attribute NewEnum.VB_UserMemId = -4
Attribute NewEnum.VB_MemberFlags = "40"
   Set NewEnum = m_Procs.[_NewEnum]
End Function

Public Function Remove(Optional ByVal Index As Long, Optional ByVal ID As Long)
   ' Check if we got a process ID.
   If ID <> 0 Then Index = IDtoIndex(ID)
   ' Return corresponding process
   If Index >= 1 And Index <= m_Count Then
      m_Procs(Index).Terminate
      Me.Refresh
   End If
End Function

' ************************************************
'  Other Public Methods
' ************************************************
Public Sub Refresh()
   Dim BytesNeeded As Long
   Dim BufferSize As Long
   Dim nRet As Long
   Dim proc As CProcess
   Dim i As Long
   
   If m_RunningNT Then
      ' Get ID list of running processes.
      BufferSize = BytesPerLong * UBound(m_PIDs)
      nRet = EnumProcesses(m_PIDs(1), BufferSize, BytesNeeded)
      If nRet = False Then
         ' Check to see if we need to allocate
         ' more space for results.
         If BytesNeeded > BufferSize Then
            ReDim m_PIDs(1 To BytesNeeded \ BytesPerLong) As Long
            BufferSize = BytesNeeded
            nRet = EnumProcesses(m_PIDs(1), BufferSize, BytesNeeded)
         End If
         ' Test for major failure, and bail if need be.
         If nRet = False Then
            m_Count = 0
            ApiErrorDump Err.LastDllError, "EnumProcesses"
            Exit Sub
         End If
      End If
      
      ' Cache number of "good" PIDs.
      m_Count = BytesNeeded \ BytesPerLong
      
      ' Fill collection with current processes.
      Set m_Procs = New Collection
      For i = 1 To m_Count
         Set proc = New CProcess
         proc.ProcessID = m_PIDs(i)
         m_Procs.Add proc, CStr(proc.ProcessID)
      Next i
      
      ' Adjust number of "good" PIDs to
      ' what is stored in collection.
      m_Count = m_Procs.Count
   End If
End Sub

' ************************************************
'  Private Methods
' ************************************************
Private Sub ApiErrorDump(ByVal ErrNum As Long, Optional ByVal Source As String = "")
   If Len(Source) Then
      Debug.Print Trim$(Source);
   Else
      Debug.Print "API";
   End If
   Debug.Print " error:"; ErrNum; " -- ";
   Debug.Print ApiErrorText(ErrNum);
End Sub

Private Function ApiErrorText(ByVal ErrNum As Long) As String
   Dim msg As String
   Dim nRet As Long

   msg = Space$(1024)
   nRet = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, ByVal 0&, ErrNum, 0&, msg, Len(msg), ByVal 0&)
   If nRet Then
      ApiErrorText = Left$(msg, nRet)
   Else
      ApiErrorText = "Error (" & ErrNum & ") not defined."
   End If
End Function

Private Function IDtoIndex(ByVal ID As Long) As Long
   Dim proc As CProcess
   Dim i As Long
   ' Convert ProcessID to collection index.
   For i = 1 To m_Procs.Count
      Set proc = m_Procs(i)
      If proc.ProcessID = ID Then
         IDtoIndex = i
         Exit For
      End If
   Next i
End Function
