Attribute VB_Name = "MPasswordSniffer"
' *********************************************************************
'  Copyright 1999-2001, Karl E. Peterson, All Rights Reserved.
'  http://www.mvps.org/vb
' *********************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code without prior written consent.
' *********************************************************************
'  Algorithm inspired by Jonathan Holmes' solution, originally written
'  with Delphi, I mainly just ported it to VB and added an interface.
' *********************************************************************
Option Explicit

Private Declare Function EnumWindows Lib "user32" (ByVal lpEnumFunc As Long, ByVal lParam As Long) As Long
Private Declare Function EnumChildWindows Lib "user32" (ByVal hWndParent As Long, ByVal lpEnumFunc As Long, ByVal lParam As Long) As Long
Private Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long) As Long
Private Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Private Const GWL_STYLE = (-16)
Private Const ES_PASSWORD = &H20&
Private Const EM_GETPASSWORDCHAR = &HD2
Private Const WM_GETTEXT = &HD
Private Const WM_GETTEXTLENGTH = &HE

Private m_Count As Long
Private m_Buffer As String
Private m_Results As String

Public Function SniffPasswords(Results As String) As Long
   ' reset counters and buffer...
   m_Count = 0
   m_Results = ""
   m_Buffer = Space$(1024)
   
   ' begin enumeration...
   Call EnumWindows(AddressOf EnumMain, 0&)
   Results = m_Results
   SniffPasswords = m_Count
End Function

Private Function EnumMain(ByVal hWnd As Long, ByVal lParam As Long) As Long
   ' stash handle and enumerate children...
   Call EnumChildWindows(hWnd, AddressOf EnumChild, hWnd)
   
   ' continue top-level enumeration...
   EnumMain = True
End Function

Private Function EnumChild(ByVal hWnd As Long, ByVal lParam As Long) As Long
   ' check for ES_PASSWORD style...
   If (GetWindowLong(hWnd, GWL_STYLE) And ES_PASSWORD) Then
      ' check if it's passworded...
      If SendMessage(hWnd, EM_GETPASSWORDCHAR, 0&, ByVal 0&) Then
         ' get top-level window text...
         m_Results = m_Results & "Caption: "
         If SendMessage(lParam, WM_GETTEXT, Len(m_Buffer), ByVal m_Buffer) Then
            m_Results = m_Results & TrimNull(m_Buffer) & vbCrLf
         Else
            m_Results = m_Results & "[None]" & vbCrLf
         End If
         
         ' retrieve password text...
         m_Results = m_Results & "   Password: "
         If SendMessage(hWnd, WM_GETTEXT, Len(m_Buffer), ByVal m_Buffer) Then
            m_Results = m_Results & TrimNull(m_Buffer) & vbCrLf & vbCrLf
         Else
            m_Results = m_Results & "[Blank]" & vbCrLf & vbCrLf
         End If
         m_Count = m_Count + 1
      End If
   End If
   
   ' continue enumeration of children...
   EnumChild = True
End Function

Private Function TrimNull(ByVal StrIn As String) As String
   Dim nul As Long
   
   ' Truncate input string at first null.
   ' If no nulls, perform ordinary Trim.
   nul = InStr(StrIn, vbNullChar)
   Select Case nul
      Case Is > 1
         TrimNull = Left$(StrIn, nul - 1)
      Case 1
         TrimNull = ""
      Case 0
         TrimNull = Trim$(StrIn)
   End Select
End Function

