Attribute VB_Name = "MTermProc"
' *************************************************************************
'  Copyright 2000-01 Karl E. Peterson
'  All Rights Reserved, http://www.mvps.org/vb
' *************************************************************************
'  You are free to use this code within your own applications, but you
'  are expressly forbidden from selling or otherwise distributing this
'  source code, non-compiled, without prior written consent.
' *************************************************************************
Option Explicit

' Win32 APIs...
Private Declare Function GetClassname Lib "user32" Alias "GetClassNameA" (ByVal hWnd As Long, ByVal lpClassName As String, ByVal nMaxCount As Long) As Long
Private Declare Function EnumWindows Lib "user32" (ByVal lpEnumFunc As Long, lParam As Any) As Long
Private Declare Function GetWindowThreadProcessId Lib "user32" (ByVal hWnd As Long, lpdwProcessId As Long) As Long
Private Declare Function GetWindowLong Lib "user32" Alias "GetWindowLongA" (ByVal hWnd As Long, ByVal nIndex As Long) As Long
Private Declare Function PostMessage Lib "user32" Alias "PostMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long
Private Declare Function SetPriorityClass Lib "kernel32" (ByVal hProcess As Long, ByVal dwPriorityClass As Long) As Long
Private Declare Function GetPriorityClass Lib "kernel32" (ByVal hProcess As Long) As Long
Private Declare Function OpenProcess Lib "kernel32" (ByVal dwDesiredAccess As Long, ByVal bInheritHandle As Long, ByVal dwProcessID As Long) As Long
Private Declare Function WaitForSingleObject Lib "kernel32" (ByVal hHandle As Long, ByVal dwMilliseconds As Long) As Long
Private Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long
Private Declare Function TerminateProcess Lib "kernel32" (ByVal hProcess As Long, ByVal uExitCode As Long) As Long
Private Declare Function IsWindowVisible Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Sub Sleep Lib "kernel32" (ByVal dwMilliseconds As Long)

' PSAPI (and related) Declares...
Private Declare Function EnumProcessModules Lib "PSAPI" (ByVal hProcess As Long, lphModule As Any, cb As Long, lpcbNeeded As Long) As Long
Private Declare Function GetModuleBaseName Lib "PSAPI" Alias "GetModuleBaseNameA" (ByVal hProcess As Long, ByVal hModule As Long, ByVal lpBasename As String, nSize As Long) As Long
Private Declare Function GetModuleFileNameEx Lib "PSAPI" Alias "GetModuleFileNameExA" (ByVal hProcess As Long, ByVal hModule As Long, ByVal lpFileName As String, nSize As Long) As Long

' INT WINAPI VDMEnumProcessWOW( PROCESSENUMPROC fp, LPARAM lparam );
Private Declare Function VDMEnumProcessWOW Lib "VDMDBG" (ByVal lpProcessNumProc As Long, ByVal lParam As Long) As Long
' INT WINAPI VDMEnumTaskWOWEx( DWORD dwProcessId, TASKENUMPROCEX fp, LPARAM lparam );
Private Declare Function VDMEnumTaskWOWEx Lib "VDMDBG" (ByVal dwProcessID As Long, ByVal lpTaskNumProcEx As Long, lParam As Any) As Long
' BOOL WINAPI VDMTerminateTaskWOW( DWORD dwProcessId, WORD htask );
Private Declare Function VDMTerminateTaskWOW Lib "VDMDBG" (ByVal dwProcessID As Long, ByVal hTask As Integer) As Long

Private Const WM_CLOSE = &H10

Private Const WAIT_FAILED = -1&        'Error on call
Private Const WAIT_OBJECT_0 = 0        'Normal completion
Private Const WAIT_TIMEOUT = &H102&    'Timeout period elapsed
Private Const STILL_ACTIVE = &H103

Private Const IGNORE = 0               'Ignore signal
Private Const INFINITE = -1&           'Infinite timeout

Private Const GWL_HWNDPARENT = (-8)

' Used by the OpenProcess API call
Private Const SYNCHRONIZE = &H100000
Private Const PROCESS_ALL_ACCESS As Long = &H1F0FFF
Private Const PROCESS_CREATE_PROCESS As Long = &H80
Private Const PROCESS_CREATE_THREAD As Long = &H2
Private Const PROCESS_DUP_HANDLE As Long = &H40
Private Const PROCESS_HEAP_ENTRY_BUSY As Long = &H4
Private Const PROCESS_HEAP_ENTRY_DDESHARE As Long = &H20
Private Const PROCESS_HEAP_ENTRY_MOVEABLE As Long = &H10
Private Const PROCESS_HEAP_REGION As Long = &H1
Private Const PROCESS_HEAP_UNCOMMITTED_RANGE As Long = &H2
Private Const PROCESS_QUERY_INFORMATION As Long = &H400
Private Const PROCESS_SET_INFORMATION As Long = &H200
Private Const PROCESS_SET_QUOTA As Long = &H100
Private Const PROCESS_TERMINATE As Long = &H1
Private Const PROCESS_VM_OPERATION As Long = &H8
Private Const PROCESS_VM_READ As Long = &H10
Private Const PROCESS_VM_WRITE As Long = &H20

' Used by SetPriorityClass
Private Const NORMAL_PRIORITY_CLASS = &H20
Private Const IDLE_PRIORITY_CLASS = &H40
Private Const HIGH_PRIORITY_CLASS = &H80
Private Const REALTIME_PRIORITY_CLASS = &H100

Public Enum ProcessPriorities
   Idle = IDLE_PRIORITY_CLASS
   Normal = NORMAL_PRIORITY_CLASS
   High = HIGH_PRIORITY_CLASS
   Realtime = REALTIME_PRIORITY_CLASS
End Enum

' Passed to callback functions
Private Type TaskInfo
   ProcessID As Long
   ThreadID As Long
   hTask16 As Integer
   MsgBoxUp As Boolean
End Type

' Used to get API error descriptions
Private Declare Function FormatMessage Lib "kernel32" Alias "FormatMessageA" _
   (ByVal dwFlags As Long, lpSource As Any, ByVal dwMessageId As Long, _
   ByVal dwLanguageId As Long, ByVal lpBuffer As String, ByVal nSize As Long, _
   Arguments As Long) As Long
Private Const FORMAT_MESSAGE_FROM_SYSTEM As Long = &H1000

' *************************************************************************
'  The methods used in this module are based loosely on the strategy
'  outlined in the following KB articles:
'    Q178893 - HOWTO: Terminate an Application "Cleanly" in Win32
'    http://support.microsoft.com/support/kb/articles/Q178/8/93.asp
' *************************************************************************
'  More info on VDMDBG:
'    Q182559 - HOWTO: Use VDMDBG Functions on Windows NT
'    http://support.microsoft.com/support/kb/articles/Q182/5/59.asp
' ************************************************************************

' Member variables
Private m_Frm As FNotResponding
Private m_FoundVdm As Boolean

' *************************************************
'  Public Methods
' *************************************************
Public Function IsProcVdmWin16(ByVal ProcessID As Long) As Boolean
   ' Clear flag, and start enumeration.
   m_FoundVdm = False
   Call VDMEnumProcessWOW(AddressOf ProcessFindMatch, ProcessID)
   IsProcVdmWin16 = m_FoundVdm
End Function

Public Function IsProcVdmDOS(ByVal ProcessID As Long) As Boolean
   Dim hProcess As Long
   Dim BufferSize As Long
   Dim BytesNeeded As Long
   Dim Modules() As Long
   Dim sBuffer As String
   Dim nRet As Long
   Dim IsVDM As Boolean
   Const opFlags As Long = _
      PROCESS_QUERY_INFORMATION Or PROCESS_VM_READ
   Const nMaxMods As Long = 256
   Const nBaseModule As Long = 1
   Const BytesPerLong As Long = 4
   Const MAX_PATH As Long = 260

   ' Open process so we can read some module info.
   hProcess = OpenProcess(opFlags, False, ProcessID)
   
   If hProcess Then
      ' Get list of process modules.
      ReDim Modules(1 To nMaxMods) As Long
      BufferSize = UBound(Modules) * BytesPerLong
      nRet = EnumProcessModules(hProcess, Modules(nBaseModule), BufferSize, BytesNeeded)
      If nRet = False Then
         ' Check to see if we need to allocate
         ' more space for results.
         If BytesNeeded > BufferSize Then
            ReDim m_Mods(nBaseModule To BytesNeeded \ BytesPerLong) As Long
            BufferSize = BytesNeeded
            nRet = EnumProcessModules(hProcess, Modules(nBaseModule), BufferSize, BytesNeeded)
         End If
      End If

      ' Snag module name.
      sBuffer = String$(MAX_PATH, 0)
      nRet = GetModuleBaseName(hProcess, Modules(nBaseModule), sBuffer, MAX_PATH)
      If nRet Then
         If UCase$(Left$(sBuffer, nRet)) = "NTVDM.EXE" Then
            ' Definitely a VDM!
            IsVDM = True
         End If
      End If
      
      ' Clean up
      Call CloseHandle(hProcess)
   End If
   
   ' Check VDMDBG enumeration to rule out
   ' that this is a Win16 VDM.
   If IsVDM Then
      IsProcVdmDOS = Not IsProcVdmWin16(ProcessID)
   End If
End Function

Public Function ProcessClose(ByVal ProcessID As Long, Optional ByVal TimeOut As Long = 5000, Optional Caption As String, Optional hIcon As Long) As Boolean
   ' First step is to post a WM_CLOSE to all
   ' top-level windows owned by process.
   Call EnumWindows(AddressOf PostCloseMessage, ByVal ProcessID)
   
   ' Check if process is still there.  Allow (default)
   ' five seconds for dialogs to be responded to.
   If WaitForProcessClose(ProcessID, TimeOut) Then
      ' Report success.
      ProcessClose = True
   
   Else
      ' Okay, now things get trickier!
      ' Setup and show "Not responding" form.
      Set m_Frm = New FNotResponding
      m_Frm.TaskName = Caption
      m_Frm.hIcon = hIcon
      m_Frm.Show
      
      ' Loop until "Not responding" form is hidden,
      ' which happens when either button is pressed.
      Do While IsWindowVisible(m_Frm.hWnd)
         ' Yield execution.
         DoEvents
         ' Need to monitor process to see if it closes
         ' on its own (in response to user input,
         ' perhaps) while dialog is up.
         If WaitForProcessClose(ProcessID, 10) = True Then
            ' Process has had normal shutdown!
            m_Frm.Cancelled = True
            ProcessClose = True
         End If
      Loop
      
      ' If user said to "End Now!", kill that sucker.
      If m_Frm.Cancelled = False Then
         ProcessClose = ProcessKill(ProcessID)
      End If
      
      ' Clean up dialog.
      Unload m_Frm
      Set m_Frm = Nothing
   End If
End Function

Public Function ProcessClose16(ByVal hWnd As Long, Optional ByVal TimeOut As Long = 10000, Optional Caption As String, Optional hIcon As Long) As Boolean
   Dim ti As TaskInfo
   Const FirstDelay As Long = 100   'milliseconds
   Const SecondDelay As Long = 1000 'milliseconds
   Const VisibleDelay As Long = 15  'seconds
   
   ' First step is to post a WM_CLOSE to all
   ' top-level windows owned by process.
   ti.ThreadID = GetWindowThreadProcessId(hWnd, ti.ProcessID)
   Call EnumWindows(AddressOf PostCloseMessage16, ti)
   
   ' No wait function for 16-bit WOW tasks in NTVDM,
   ' so all we can do is wait.  First, just give the
   ' task a quick (100ms) to see if the WM_CLOSE is
   ' acted upon.  Then, if not, wait (default: 10s)
   ' for awhile to see if user responds to possible
   ' dialogs.
   Call Sleep(FirstDelay)
   If GetWindowThreadProcessId(hWnd, ti.ProcessID) Then
      ' Guess we're gonna hafta play a bit rough!
      Call Sleep(TimeOut)
      ' Kill it. No worries, if already dead.
      ' Enumerate tasks within this VDM.
      Call VDMEnumTaskWOWEx(ti.ProcessID, AddressOf TaskEnumProcEx, ti)
      If ti.hTask16 Then
         ' Look for msgbox...
         Call EnumWindows(AddressOf CheckMsgBox16, ti)
         If ti.MsgBoxUp Then
            ' Setup and show "Not responding" form.
            Set m_Frm = New FNotResponding
            m_Frm.TaskName = Caption
            m_Frm.hIcon = hIcon
            m_Frm.Prompt = Win16Waiting
            m_Frm.Show
            
            ' Loop until "Not responding" form is hidden,
            ' which happens when either button is pressed.
            Do While IsWindowVisible(m_Frm.hWnd)
               ' Yield execution.
               DoEvents
               ' Need to monitor process to see if it
               ' its messagebox is closed.
               ti.MsgBoxUp = False
               Call EnumWindows(AddressOf CheckMsgBox16, ti)
               If ti.MsgBoxUp = False Then
                  ' Task is no longer waiting, give
                  ' it a chance to shutdown.
                  m_Frm.Cancelled = True
                  Call Sleep(SecondDelay)
               End If
            Loop
            Unload m_Frm
            
            ' If user said to "End Now!", kill that sucker.
            If m_Frm.Cancelled = False Then
               Call VDMTerminateTaskWOW(ti.ProcessID, ti.hTask16)
            
            Else
               ' Get ready to display countdown form.
               Set m_Frm = New FNotResponding
               m_Frm.TaskName = Caption
               m_Frm.hIcon = hIcon
               m_Frm.Prompt = Win16NotResponding
               m_Frm.Wait = VisibleDelay
               m_Frm.Show
               Do While GetWindowThreadProcessId(hWnd, ti.ProcessID)
                  ' Yield execution.
                  DoEvents
                  ' If form is gone, bail from loop.
                  If IsWindowVisible(m_Frm.hWnd) = False Then
                     Exit Do
                  End If
               Loop
               
               ' User either said to "End Now!", or
               ' timeout expired - kill that sucker.
               If m_Frm.Cancelled = False Then
                  Call VDMTerminateTaskWOW(ti.ProcessID, ti.hTask16)
               End If
               Unload m_Frm
            End If
         End If
      End If
   End If
   
   ' Clean up dialog.
   If Not m_Frm Is Nothing Then
      Set m_Frm = Nothing
   End If
End Function

Public Function ProcessCloseDOS(ByVal ProcessID As Long) As Boolean
   ' Post WM_CLOSE to all top-level windows
   ' owned by process. From that point on,
   ' NT takes over and shows the End Now
   ' dialog if the process is non-responsive.
   Call EnumWindows(AddressOf PostCloseMessage, ByVal ProcessID)
End Function

Public Function ProcessKill(ByVal ProcessID As Long) As Boolean
   Dim hProc As Long
   Const fdwAccess As Long = SYNCHRONIZE Or PROCESS_TERMINATE
   
   ' Need to open process with terminate rights, or
   ' just give up immediately.
   hProc = OpenProcess(fdwAccess, 0&, ProcessID)
   If hProc Then
      ' Kill it.
      If TerminateProcess(hProc, 0&) Then
         ProcessKill = True
      Else
         ApiErrorDump Err.LastDllError, "TerminateProcess"
      End If
      ' Clean up.
      Call CloseHandle(hProc)
   End If
End Function

Public Function ProcessPrioritySet(Optional ByVal ProcessID As Long, Optional ByVal hWnd As Long, Optional ByVal Priority As ProcessPriorities = NORMAL_PRIORITY_CLASS) As Long
   Dim hProc As Long
   Const fdwAccess1 As Long = PROCESS_QUERY_INFORMATION Or PROCESS_SET_INFORMATION
   Const fdwAccess2 As Long = PROCESS_QUERY_INFORMATION
   
   ' If not passed a PID, then find value for hWnd.
   If ProcessID = 0 Then
      Call GetWindowThreadProcessId(hWnd, ProcessID)
   End If
   
   ' Need to open process with setinfo rights.
   hProc = OpenProcess(fdwAccess1, 0&, ProcessID)
   If hProc Then
      ' Attempt to set new priority.
      Call SetPriorityClass(hProc, Priority)
   Else
      ' Weren't allowed to setinfo, so just open to
      ' enable return of current priority setting.
      hProc = OpenProcess(fdwAccess2, 0&, ProcessID)
   End If
   
   ' Get current/new setting.
   ProcessPrioritySet = GetPriorityClass(hProc)
   ' Clean up.
   Call CloseHandle(hProc)
End Function

Public Function WaitForProcessClose(ByVal PID As Long, ByVal TimeOut As Long) As Boolean
   ' TimeOut needs to be passed in milliseconds!
   Dim hProc As Long
   Dim nRet As Long
   Const fdwAccess = SYNCHRONIZE
   
   ' Attempt to open process, and wait on it to close.
   hProc = OpenProcess(fdwAccess, False, PID)
   
   If hProc Then
      nRet = WaitForSingleObject(hProc, TimeOut)
      Select Case nRet
         Case WAIT_TIMEOUT
            ' Still open.
            WaitForProcessClose = False
         Case WAIT_OBJECT_0
            ' Process has closed.
            WaitForProcessClose = True
         Case Else
            ' Error on call.
            WaitForProcessClose = False
            ApiErrorDump Err.LastDllError, "WaitForSingleObject"
      End Select
   Else
      ' Nothing we can do, at this point.
      ' Process may be gone already?
      ApiErrorDump Err.LastDllError, "OpenProcess"
      WaitForProcessClose = True
   End If
   ' Clean up.
   Call CloseHandle(hProc)
End Function

' ************************************************************
'  Private methods
' ************************************************************
Private Sub ApiErrorDump(ByVal ErrNum As Long, Optional ByVal Source As String = "")
   If Len(Source) Then
      Debug.Print Trim$(Source);
   Else
      Debug.Print "API";
   End If
   Debug.Print " error:"; ErrNum; " -- ";
   Debug.Print ApiErrorText(ErrNum);
End Sub

Private Function ApiErrorText(ByVal ErrNum As Long) As String
   Dim msg As String
   Dim nRet As Long

   msg = Space$(1024)
   nRet = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, ByVal 0&, ErrNum, 0&, msg, Len(msg), ByVal 0&)
   If nRet Then
      ApiErrorText = Left$(msg, nRet)
   Else
      ApiErrorText = "Error (" & ErrNum & ") not defined."
   End If
End Function

Private Function CheckMsgBox16(ByVal hWnd As Long, lParam As TaskInfo) As Long
   Static PID As Long
   Static ThreadID As Long
   ' Default to continued enumeration.
   CheckMsgBox16 = True
   ' Check if hWnd belongs to PID: lParam
   ThreadID = GetWindowThreadProcessId(hWnd, PID)
   If PID = lParam.ProcessID Then
      ' Right process, check for right thread (task).
      If ThreadID = lParam.ThreadID Then
         ' Is this a MsgBox?
         If Classname(hWnd) = "#32770" Then
            lParam.MsgBoxUp = True
            ' Stop enumeration
            CheckMsgBox16 = False
         End If
      End If
   End If
End Function

Private Function Classname(ByVal hWnd As Long) As String
   Dim nRet As Long
   Dim Class As String
   Const MaxLen As Long = 256
   
   ' Retrieve classname of passed window.
   Class = String$(MaxLen, 0)
   nRet = GetClassname(hWnd, Class, MaxLen)
   If nRet Then Classname = Left$(Class, nRet)
End Function

' BOOL WINAPI ProcessEnumProc( DWORD dwProcessId, DWORD dwAttrib, LPARAM t )
Private Function ProcessFindMatch(ByVal dwProcessID As Long, ByVal dwAttrib As Long, ByVal lParam As Long) As Long
   If dwProcessID = lParam Then m_FoundVdm = True
   ' Return TRUE to stop enumeration or FALSE to continue enumeration
   ProcessFindMatch = m_FoundVdm
End Function

Private Function PostCloseMessage(ByVal hWnd As Long, ByVal lParam As Long) As Long
   Static PID As Long
   ' Check if hWnd belongs to PID: lParam
   Call GetWindowThreadProcessId(hWnd, PID)
   If PID = lParam Then
      ' Belongs to target process, try close message
      ' if this doesn't have an owner. VB apps that
      ' display a msgbox from QueryUnload will lock
      ' a 9x box up hard, if we don't do this test!
      If GetWindowLong(hWnd, GWL_HWNDPARENT) = 0 Then
         Call PostMessage(hWnd, WM_CLOSE, 0&, ByVal 0&)
      End If
   End If
   ' Return true to continue enumeration.
   PostCloseMessage = True
End Function

Private Function PostCloseMessage16(ByVal hWnd As Long, lParam As TaskInfo) As Long
   Static PID As Long
   Static ThreadID As Long
   ' Check if hWnd belongs to PID: lParam
   ThreadID = GetWindowThreadProcessId(hWnd, PID)
   If PID = lParam.ProcessID Then
      ' Right process, check for right thread (task).
      If ThreadID = lParam.ThreadID Then
         ' Belongs to target task, try close message
         ' if this doesn't have an owner. VB apps that
         ' display a msgbox from QueryUnload will lock
         ' a 9x box up hard, if we don't do this test!
         If GetWindowLong(hWnd, GWL_HWNDPARENT) = 0 Then
            Call PostMessage(hWnd, WM_CLOSE, 0&, ByVal 0&)
         End If
      End If
   End If
   ' Return true to continue enumeration.
   PostCloseMessage16 = True
End Function

' BOOL WINAPI TaskEnumProcEx( DWORD dwThreadId, WORD hMod16, WORD hTask16, PSZ pszModName, PSZ pszFileName, LPARAM lParam )
Private Function TaskEnumProcEx(ByVal dwThreadId As Long, ByVal hMod16 As Integer, ByVal hTask16 As Integer, ByVal lpszModName As Long, ByVal lpszFilename As Long, lParam As TaskInfo) As Long
   If dwThreadId = lParam.ThreadID Then
      ' Found the one we're after.
      ' Cache hTask16 and stop enum.
      lParam.hTask16 = hTask16
      TaskEnumProcEx = True
   End If
End Function


